# :coding: utf-8
# :copyright: Copyright (c) 2013 Martin Pengelly-Phillips
# :license: See LICENSE.txt.
from __future__ import absolute_import

import os
import uuid
import imp

from lucidity._version import __version__
from lucidity.template import Template, Resolver
from lucidity.error import ParseError, FormatError, NotFound


def discover_templates(paths=None, recursive=True):
    '''Search *paths* for mount points and load templates from them.

    *paths* should be a list of filesystem paths to search for mount points.
    If not specified will try to use value from environment variable
    :envvar:`LUCIDITY_TEMPLATE_PATH`.

    A mount point is a Python file that defines a 'register' function. The
    function should return a list of instantiated
    :py:class:`~lucidity.template.Template` objects.

    If *recursive* is True (the default) then all directories under a path
    will also be searched.

    '''
    templates = []

    if paths is None:
        paths = os.environ.get('LUCIDITY_TEMPLATE_PATH', '').split(os.pathsep)

    for path in paths:
        for base, directories, filenames in os.walk(path):
            for filename in filenames:
                _, extension = os.path.splitext(filename)
                if extension != '.py':
                    continue

                module_path = os.path.join(base, filename)
                module_name = uuid.uuid4().hex
                module = imp.load_source(module_name, module_path)
                try:
                    registered = module.register()
                except AttributeError:
                    pass
                else:
                    if registered:
                        templates.extend(registered)

            if not recursive:
                del directories[:]

    return templates


def parse(path, templates):
    '''Parse *path* against *templates*.

    *path* should be a string to parse.

    *templates* should be a list of :py:class:`~lucidity.template.Template`
    instances in the order that they should be tried.

    Return ``(data, template)`` from first successful parse.

    Raise :py:class:`~lucidity.error.ParseError` if *path* is not
    parseable by any of the supplied *templates*.

    '''
    for template in templates:
        try:
            data = template.parse(path)
        except ParseError:
            continue
        else:
            return (data, template)

    raise ParseError(
        'Path {0!r} did not match any of the supplied template patterns.'
        .format(path)
    )


def format(data, templates):  # @ReservedAssignment
    '''Format *data* using *templates*.

    *data* should be a dictionary of data to format into a path.

    *templates* should be a list of :py:class:`~lucidity.template.Template`
    instances in the order that they should be tried.

    Return ``(path, template)`` from first successful format.

    Raise :py:class:`~lucidity.error.FormatError` if *data* is not
    formattable by any of the supplied *templates*.


    '''
    for template in templates:
        try:
            path = template.format(data)
        except FormatError:
            continue
        else:
            return (path, template)

    raise FormatError(
        'Data {0!r} was not formattable by any of the supplied templates.'
        .format(data)
    )


def get_template(name, templates):
    '''Retrieve a template from *templates* by *name*.

    Raise :py:exc:`~lucidity.error.NotFound` if no matching template with
    *name* found in *templates*.

    '''
    for template in templates:
        if template.name == name:
            return template

    raise NotFound(
        '{0} template not found in specified templates.'.format(name)
    )



"""
This file is part of SPIL, The Simple Pipeline Lib.

(C) copyright 2019-2023 Michael Haussmann, spil@xeo.info

SPIL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.

SPIL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with SPIL.
If not, see <https://www.gnu.org/licenses/>.
"""
from __future__ import annotations
from typing import Iterator, List, overload, Optional

from typing_extensions import Literal

import itertools as it

from spil import Sid
from spil.sid.read.finder import Finder

from spil.util.log import debug, warning, error


class FindByGlob(Finder):
    """
    Parent class for glob type searches:
    FindInFiles, FindInList, FindInConstants

    The search process is as follows:
        find():
        - the search sid string is "unfolded" into a list of typed search Sids.

        do_find()
        - depending on the types of searches, defined by the search symbols ('>', ...), the search is delegated to a finder function.
        (currently either "sorted_search" or "star_search").
    """

    @overload
    def do_find(self, search_sids: List[Sid], as_sid: Literal[True]) -> Iterator[Sid]:  # noqa
        ...

    @overload
    def do_find(self, search_sids: List[Sid], as_sid: Literal[False]) -> Iterator[str]:  # noqa
        ...

    @overload
    def do_find(  # noqa
        self, search_sids: List[Sid], as_sid: Optional[bool]
    ) -> Iterator[Sid] | Iterator[str]:
        ...

    def do_find(self, search_sids, as_sid=True):

        # depending on input, select the right generator
        is_sorted_search = any([ssid.string.count(">") for ssid in search_sids])

        if not search_sids:
            warning("Nothing Searchable. ")
            generator = ()
        elif is_sorted_search:
            generator = self.sorted_search(search_sids, as_sid=as_sid)
        else:
            generator = self.star_search(search_sids, as_sid=as_sid)

        yield from generator

    @overload
    def star_search(
        self, search_sids: List[Sid], as_sid: Literal[False], do_sort: bool = False
    ) -> Iterator[str]:
        ...

    @overload
    def star_search(
        self, search_sids: List[Sid], as_sid: Literal[True], do_sort: bool = False
    ) -> Iterator[Sid]:
        ...

    @overload
    def star_search(
        self, search_sids: List[Sid], as_sid: Optional[bool], do_sort: bool
    ) -> Iterator[str] | Iterator[Sid]:
        ...

    def star_search(self, search_sids, as_sid=False, do_sort=False):
        msg = f'Current instance "{self}" has no star_search. FindByGlob is abstract, use FindInFiles or FindInList.'
        raise NotImplementedError(msg)

    def sorted_search(
        self, search_sids: List[Sid], as_sid: bool = True
    ) -> Iterator[Sid] | Iterator[str]:
        """
        Operates a sorted read.
        A sorted read contains the ">" sign, standing for "last"
        or the "<" sign, standing for "first". (not yet implemented)

        TODO: "meaningful sort" (eg. LAY < ANI < RND), currently only alphanumerical sort.

        :param search_sids:
        :param as_sid:
        :return:
        """
        # index is coherent in all search_sids, which is a bit strange
        index = str(search_sids[0]).split("/").index(">")

        """
        indices = []
        for ss in search_sids:
            indices.append(str(ss).split('/').index('>'))
        print('ind' + str(indices))
        indices = list(set(indices))
        """
        # indices = [i for i, x in enumerate(str(search_sid).split('/')) if x == '>']
        # debug index, indices

        founds: List[str] = []
        for search_sid in search_sids:
            ssid = search_sid.uri.replace(">", "*")
            debug("star read start on {}".format(ssid))
            founds.extend(self.star_search([Sid(ssid)], as_sid=False))
            debug("star read done")

        founds = sorted(list(set(founds)), reverse=True)
        # TODO: sort by row - and resort after each narrowing
        # pprint(founds)
        debug("found {} matches".format(len(founds)))

        # for index in indices:
        # works with > in any position, but does not use sort by row, and no delegate sorting, and no special sorting
        for key, grp in it.groupby(founds, key=lambda x: x.split("/")[0:index]):
            result = list(grp)
            # debug('{}: {}'.format(key, result))
            if as_sid:
                yield Sid(result[0])
            else:
                yield result[0]

        """
        for index in indices:
            if not founds:
                break
            filtered = []
            for key, grp in it.groupby(founds, key=lambda x: x.split('/')[0:index]):
                result = list(grp)
                debug('{}: "{}" {}'.format(key, result[0].split('/')[index], result))
                filtered.append(result[0])
                filtered.extend(result)
            founds = filtered
        """

        """
        for index in indices:
            for key, grp in it.groupby(founds, key=lambda x: x.split('/')[0:index]):
                result = list(grp)
                debug('{}: "{}" {}'.format(key, result[0].split('/')[index], result))
        """

    """
    Problem:

    FTOT/S/SQ0001/SH0020/**/cache,maya?state=WIP&version=>
    Doesn't return FTOT/S/SQ0001/SH0020/ANI/V019/WIP/CAM/abc
    See tests/filesearch.

    """


if __name__ == "__main__":

    pass



# -*- coding: utf-8 -*-
"""
This file is part of SPIL, The Simple Pipeline Lib.

(C) copyright 2019-2023 Michael Haussmann, spil@xeo.info

SPIL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.

SPIL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with SPIL.
If not, see <https://www.gnu.org/licenses/>.
"""
from __future__ import annotations
from typing import Iterable, List, Set, Optional, Mapping, Any


# FIXME: This is work in progress.

from spil import Sid, Writer
# from spil.conf import get_creator, get_updater, get_setter, get_deleter


class WriteToAll(Writer):
    """
    This Writer is a thin layer to delegate all Write operations to configured functions.
    This allows existing or external pipeline write functions to be wrapped by a Writer, without the need to implement one.

    Note that it is recommended to implement a custom writer, because the code is better structured,
    but this approach is not always possible or desirable.
    """

    def create(self, sid: Sid | str, data: Optional[dict] = None) -> bool:
        """
        Creates the given Sid, optionally with given data.

        Args:
            sid: Sid string or instance
            data: optional data dictionary

        Returns: True on success, False on failure
        """
        return False

    def update(self, sid: Sid | str, data: Mapping[str, Any]) -> bool:
        """
        Updates given data on given Sid.

        Args:
            sid: Sid string or instance
            data: data dictionary

        Returns: True on success, False on failure
        """
        return False

    def set(self, sid: Sid | str, attribute: Optional[str], value: Optional[Any], **kwargs) -> bool:
        """
        Sets given attribute with given value, and / or given kwargs, for given Sid.

        Args:
            sid: Sid string or instance
            attribute: attribute name, data key
            value: value for data
            **kwargs: key / value pairs.

        Returns: True on success, False on failure
        """
        return False

    def delete(self, sid: Sid | str) -> bool:
        """
        Deletes given Sid.

        Args:
            sid: Sid string or instance

        Returns: True on success, False on failure
        """
        return False

    def __str__(self):
        return f"[spil.{self.__class__.__name__}]"


if __name__ == "__main__":
    print(WriteToAll())
